from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.db.models import Q
from django.utils.translation import npgettext, pgettext, pgettext_lazy

from ...acl.models import Role
from ...admin.forms import IsoDateTimeField, YesNoSwitch
from ...core.validators import validate_sluggable
from ...notifications.threads import ThreadNotifications
from ...search.filter_queryset import filter_queryset
from ..models import Ban, DataDownload, Rank
from ..profilefields import profilefields
from ..utils import hash_email, slugify_username
from ..validators import validate_email, validate_username

User = get_user_model()


class UserBaseForm(forms.ModelForm):
    username = forms.CharField(label=pgettext_lazy("admin user form", "Username"))
    title = forms.CharField(
        label=pgettext_lazy("admin user form", "Custom title"),
        required=False,
    )
    email = forms.EmailField(label=pgettext_lazy("admin user form", "E-mail address"))

    class Meta:
        model = User
        fields = ["username", "email", "title"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        self.settings = self.request.settings

        super().__init__(*args, **kwargs)

    def clean_username(self):
        data = self.cleaned_data["username"]
        validate_username(self.settings, data, exclude=self.instance)
        return data

    def clean_email(self):
        data = self.cleaned_data["email"]
        validate_email(data, exclude=self.instance)
        return data

    def clean_new_password(self):
        data = self.cleaned_data["new_password"]
        if data:
            validate_password(data, user=self.instance)
        return data

    def clean_roles(self):
        data = self.cleaned_data["roles"]

        for role in data:
            if role.special_role == "authenticated":
                break
        else:
            message = pgettext(
                "admin user form", 'All registered members must have a "Member" role.'
            )
            raise forms.ValidationError(message)

        return data


class NewUserForm(UserBaseForm):
    new_password = forms.CharField(
        label=pgettext_lazy("admin user form", "Password"),
        strip=False,
        widget=forms.PasswordInput,
    )

    class Meta:
        model = User
        fields = ["username", "email", "title"]


class EditUserForm(UserBaseForm):
    IS_STAFF_LABEL = pgettext_lazy("admin user form", "Is administrator")
    IS_STAFF_HELP_TEXT = pgettext_lazy(
        "admin user form",
        "Designates whether the user can log into admin sites. If Django admin site is enabled, this user will need additional permissions assigned within it to admin Django modules.",
    )

    IS_SUPERUSER_LABEL = pgettext_lazy("admin user form", "Is superuser")
    IS_SUPERUSER_HELP_TEXT = pgettext_lazy(
        "admin user form",
        "Only administrators can access admin sites. In addition to admin site access, superadmins can also change other members admin levels.",
    )

    IS_ACTIVE_LABEL = pgettext_lazy("admin user form", "Is active")
    IS_ACTIVE_HELP_TEXT = pgettext_lazy(
        "admin user form",
        "Designates whether this user should be treated as active. Turning this off is non-destructible way to remove user accounts.",
    )

    IS_ACTIVE_STAFF_MESSAGE_LABEL = pgettext_lazy("admin user form", "Staff message")
    IS_ACTIVE_STAFF_MESSAGE_HELP_TEXT = pgettext_lazy(
        "admin user form",
        "Optional message for forum team members explaining why user's account has been disabled.",
    )

    new_password = forms.CharField(
        label=pgettext_lazy("admin user form", "Set new password"),
        strip=False,
        widget=forms.PasswordInput,
        required=False,
    )

    is_avatar_locked = YesNoSwitch(
        label=pgettext_lazy("admin user form", "Lock avatar changes"),
        help_text=pgettext_lazy(
            "admin user form",
            "Setting this to yes will stop user from changing their avatar, and will reset their avatar to procedurally generated one.",
        ),
    )
    avatar_lock_user_message = forms.CharField(
        label=pgettext_lazy("admin user form", "User message"),
        help_text=pgettext_lazy(
            "admin user form",
            "Optional message for user explaining why they are banned form changing avatar.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )
    avatar_lock_staff_message = forms.CharField(
        label=pgettext_lazy("admin user form", "Staff message"),
        help_text=pgettext_lazy(
            "admin user form",
            "Optional message for forum team members explaining why user is banned form changing avatar.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )

    signature = forms.CharField(
        label=pgettext_lazy("admin user form", "Signature contents"),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )
    is_signature_locked = YesNoSwitch(
        label=pgettext_lazy("admin user form", "Lock signature changes"),
        help_text=pgettext_lazy(
            "admin user form",
            "Setting this to yes will stop user from making changes to his/her signature.",
        ),
    )
    signature_lock_user_message = forms.CharField(
        label=pgettext_lazy("admin user form", "User message"),
        help_text=pgettext_lazy(
            "admin user form",
            "Optional message to user explaining why his/hers signature is locked.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )
    signature_lock_staff_message = forms.CharField(
        label=pgettext_lazy("admin user form", "Staff message"),
        help_text=pgettext_lazy(
            "admin user form",
            "Optional message to team members explaining why user signature is locked.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )

    is_hiding_presence = YesNoSwitch(
        label=pgettext_lazy("admin user form", "Hides presence")
    )

    limits_private_thread_invites_to = forms.TypedChoiceField(
        label=pgettext_lazy("admin user form", "Who can add user to private threads"),
        coerce=int,
        choices=User.LIMIT_INVITES_TO_CHOICES,
    )

    subscribe_to_started_threads = forms.TypedChoiceField(
        label=pgettext_lazy("admin user form", "Started threads"),
        coerce=int,
        choices=User.SUBSCRIPTION_CHOICES,
    )
    subscribe_to_replied_threads = forms.TypedChoiceField(
        label=pgettext_lazy("admin user form", "Replied threads"),
        coerce=int,
        choices=User.SUBSCRIPTION_CHOICES,
    )

    watch_started_threads = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin user form",
            "Automatically watch threads that the user has started",
        ),
        coerce=int,
        choices=ThreadNotifications.choices,
    )
    watch_replied_threads = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin user form",
            "Automatically watch threads that the user has replied to",
        ),
        coerce=int,
        choices=ThreadNotifications.choices,
    )
    watch_new_private_threads_by_followed = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin user form",
            "Automatically watch private threads that the user was invited to by users they are following",
        ),
        coerce=int,
        choices=ThreadNotifications.choices,
    )
    watch_new_private_threads_by_other_users = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin user form",
            "Automatically watch private threads that the user was invited to by other users",
        ),
        coerce=int,
        choices=ThreadNotifications.choices,
    )
    notify_new_private_threads_by_followed = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin user form",
            "Notify about new private thread invitations from users this user is following",
        ),
        coerce=int,
        choices=ThreadNotifications.choices,
    )
    notify_new_private_threads_by_other_users = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin user form",
            "Notify about new private thread invitations from other users",
        ),
        coerce=int,
        choices=ThreadNotifications.choices,
    )

    class Meta:
        model = User
        fields = [
            "username",
            "email",
            "title",
            "is_avatar_locked",
            "avatar_lock_user_message",
            "avatar_lock_staff_message",
            "signature",
            "is_signature_locked",
            "is_hiding_presence",
            "limits_private_thread_invites_to",
            "signature_lock_user_message",
            "signature_lock_staff_message",
            "subscribe_to_started_threads",
            "subscribe_to_replied_threads",
            "watch_started_threads",
            "watch_replied_threads",
            "watch_new_private_threads_by_followed",
            "watch_new_private_threads_by_other_users",
            "notify_new_private_threads_by_followed",
            "notify_new_private_threads_by_other_users",
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        profilefields.add_fields_to_admin_form(self.request, self.instance, self)

    def get_profile_fields_groups(self):
        profile_fields_groups = []
        for group in self._profile_fields_groups:
            fields_group = {"name": group["name"], "fields": []}

            for fieldname in group["fields"]:
                fields_group["fields"].append(self[fieldname])

            profile_fields_groups.append(fields_group)
        return profile_fields_groups

    def clean_signature(self):
        data = self.cleaned_data["signature"]

        length_limit = self.settings.signature_length_max
        if len(data) > length_limit:
            message = npgettext(
                "signature length validator",
                "Signature can't be longer than %(limit)s character.",
                "Signature can't be longer than %(limit)s characters.",
                length_limit,
            )
            raise forms.ValidationError(message % {"limit": length_limit})

        return data

    def clean(self):
        data = super().clean()
        return profilefields.clean_form(self.request, self.instance, self, data)


def UserFormFactory(FormType, instance):
    extra_fields = {}

    extra_fields["rank"] = forms.ModelChoiceField(
        label=pgettext_lazy("admin user form", "Rank"),
        help_text=pgettext_lazy(
            "admin user form",
            "Ranks are used to group and distinguish users. They are also used to add permissions to groups of users.",
        ),
        queryset=Rank.objects.order_by("name"),
        initial=instance.rank,
    )

    roles = Role.objects.order_by("name")

    extra_fields["roles"] = forms.ModelMultipleChoiceField(
        label=pgettext_lazy("admin user form", "Roles"),
        help_text=pgettext_lazy(
            "admin user form",
            'Individual roles of this user. All users must have a "Member" role.',
        ),
        queryset=roles,
        initial=instance.roles.all() if instance.pk else None,
        widget=forms.CheckboxSelectMultiple,
    )

    return type("UserFormFinal", (FormType,), extra_fields)


def StaffFlagUserFormFactory(FormType, instance):
    staff_fields = {
        "is_staff": YesNoSwitch(
            label=EditUserForm.IS_STAFF_LABEL,
            help_text=EditUserForm.IS_STAFF_HELP_TEXT,
            initial=instance.is_staff,
        ),
        "is_superuser": YesNoSwitch(
            label=EditUserForm.IS_SUPERUSER_LABEL,
            help_text=EditUserForm.IS_SUPERUSER_HELP_TEXT,
            initial=instance.is_superuser,
        ),
    }

    return type("StaffUserForm", (FormType,), staff_fields)


def UserIsActiveFormFactory(FormType, instance):
    is_active_fields = {
        "is_active": YesNoSwitch(
            label=EditUserForm.IS_ACTIVE_LABEL,
            help_text=EditUserForm.IS_ACTIVE_HELP_TEXT,
            initial=instance.is_active,
        ),
        "is_active_staff_message": forms.CharField(
            label=EditUserForm.IS_ACTIVE_STAFF_MESSAGE_LABEL,
            help_text=EditUserForm.IS_ACTIVE_STAFF_MESSAGE_HELP_TEXT,
            initial=instance.is_active_staff_message,
            widget=forms.Textarea(attrs={"rows": 3}),
            required=False,
        ),
    }

    return type("UserIsActiveForm", (FormType,), is_active_fields)


def EditUserFormFactory(
    FormType, instance, add_is_active_fields=False, add_admin_fields=False
):
    FormType = UserFormFactory(FormType, instance)

    if add_is_active_fields:
        FormType = UserIsActiveFormFactory(FormType, instance)

    if add_admin_fields:
        FormType = StaffFlagUserFormFactory(FormType, instance)

    return FormType


class BaseFilterUsersForm(forms.Form):
    username = forms.CharField(
        label=pgettext_lazy("admin users filter form", "Username"),
        required=False,
    )
    email = forms.CharField(
        label=pgettext_lazy("admin users filter form", "E-mail"),
        required=False,
    )
    profilefields = forms.CharField(
        label=pgettext_lazy("admin users filter form", "Profile fields contain"),
        required=False,
    )
    is_inactive = forms.BooleanField(
        label=pgettext_lazy("admin users filter form", "Requires activation")
    )
    is_disabled = forms.BooleanField(
        label=pgettext_lazy("admin users filter form", "Account disabled")
    )
    is_staff = forms.BooleanField(
        label=pgettext_lazy("admin users filter form", "Administrator")
    )
    is_deleting_account = forms.BooleanField(
        label=pgettext_lazy("admin users filter form", "Deletes their account")
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("username"):
            queryset = filter_queryset(
                queryset, "slug", slugify_username(criteria["username"])
            )

        if criteria.get("email"):
            queryset = filter_queryset(
                queryset, "email", criteria["email"], case_sensitive=False
            )

        if criteria.get("rank"):
            queryset = queryset.filter(rank_id=criteria["rank"])

        if criteria.get("role"):
            queryset = queryset.filter(roles__id=criteria["role"])

        if criteria.get("is_inactive"):
            queryset = queryset.filter(requires_activation__gt=0)

        if criteria.get("is_disabled"):
            queryset = queryset.filter(is_active=False)

        if criteria.get("is_staff"):
            queryset = queryset.filter(is_staff=True)

        if criteria.get("is_deleting_account"):
            queryset = queryset.filter(is_deleting_account=True)

        if criteria.get("profilefields", "").strip():
            queryset = profilefields.search_users(
                criteria.get("profilefields").strip(), queryset
            )

        return queryset


def create_filter_users_form():
    """
    Factory that uses cache for ranks and roles,
    and makes those ranks and roles typed choice fields that play nice
    with passing values via GET
    """
    ranks_choices = [("", pgettext_lazy("admin users rank filter choice", "All ranks"))]
    for rank in Rank.objects.order_by("name").iterator():
        ranks_choices.append((rank.pk, str(rank)))

    roles_choices = [("", pgettext_lazy("admin users role filter choice", "All roles"))]
    for role in Role.objects.order_by("name").iterator():
        roles_choices.append((role.pk, str(role)))

    extra_fields = {
        "rank": forms.TypedChoiceField(
            label=pgettext_lazy("admin users filter form", "Has rank"),
            coerce=int,
            required=False,
            choices=ranks_choices,
        ),
        "role": forms.TypedChoiceField(
            label=pgettext_lazy("admin users filter form", "Has role"),
            coerce=int,
            required=False,
            choices=roles_choices,
        ),
    }

    return type("FilterUsersForm", (BaseFilterUsersForm,), extra_fields)


class RankForm(forms.ModelForm):
    name = forms.CharField(
        label=pgettext_lazy("admin rank form", "Name"),
        validators=[validate_sluggable()],
        help_text=pgettext_lazy(
            "admin rank form",
            'Short and descriptive name of all users with this rank. "The Team" or "Game Masters" are good examples.',
        ),
    )
    title = forms.CharField(
        label=pgettext_lazy("admin rank form", "User title"),
        required=False,
        help_text=pgettext_lazy(
            "admin rank form",
            'Optional, singular version of rank name displayed by user names. For example "GM" or "Dev".',
        ),
    )
    description = forms.CharField(
        label=pgettext_lazy("admin rank form", "Description"),
        max_length=2048,
        required=False,
        widget=forms.Textarea(attrs={"rows": 3}),
        help_text=pgettext_lazy(
            "admin rank form",
            "Optional description explaining function or status of members distincted with this rank.",
        ),
    )
    roles = forms.ModelMultipleChoiceField(
        label=pgettext_lazy("admin rank form", "User roles"),
        widget=forms.CheckboxSelectMultiple,
        queryset=Role.objects.order_by("name"),
        required=False,
        help_text=pgettext_lazy(
            "admin rank form", "Rank can give additional roles to users with it."
        ),
    )
    css_class = forms.CharField(
        label=pgettext_lazy("admin rank form", "CSS class"),
        required=False,
        help_text=pgettext_lazy(
            "admin rank form",
            "Optional css class added to content belonging to this rank owner.",
        ),
    )
    is_tab = YesNoSwitch(
        label=pgettext_lazy("admin rank form", "Give rank dedicated tab on users list"),
        required=False,
        help_text=pgettext_lazy(
            "admin rank form",
            "Selecting this option will make users with this rank easily discoverable by others through dedicated page on forum users list.",
        ),
    )

    class Meta:
        model = Rank
        fields = ["name", "description", "css_class", "title", "roles", "is_tab"]

    def clean_name(self):
        data = self.cleaned_data["name"]
        self.instance.set_name(data)

        unique_qs = Rank.objects.filter(slug=self.instance.slug)
        if self.instance.pk:
            unique_qs = unique_qs.exclude(pk=self.instance.pk)

        if unique_qs.exists():
            raise forms.ValidationError(
                pgettext(
                    "admin rank form",
                    "There's already an other rank with this name.",
                )
            )

        return data


class BanUsersForm(forms.Form):
    ban_type = forms.MultipleChoiceField(
        label=pgettext_lazy("admin ban users form", "Values to ban"),
        widget=forms.CheckboxSelectMultiple,
        choices=[],
    )
    user_message = forms.CharField(
        label=pgettext_lazy("admin ban users form", "User message"),
        required=False,
        max_length=1000,
        help_text=pgettext_lazy(
            "admin ban users form",
            "Optional message displayed to users instead of default one.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": pgettext_lazy(
                "admin ban users form", "Message can't be longer than 1000 characters."
            )
        },
    )
    staff_message = forms.CharField(
        label=pgettext_lazy("admin ban users form", "Team message"),
        required=False,
        max_length=1000,
        help_text=pgettext_lazy(
            "admin ban users form",
            "Optional ban message for moderators and administrators.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": pgettext_lazy(
                "admin ban users form", "Message can't be longer than 1000 characters."
            )
        },
    )
    expires_on = IsoDateTimeField(
        label=pgettext_lazy("admin ban users form", "Expiration date"), required=False
    )

    def __init__(self, *args, **kwargs):
        users = kwargs.pop("users")

        super().__init__(*args, **kwargs)

        self.fields["ban_type"].choices = [
            ("usernames", pgettext_lazy("admin ban users form", "Usernames")),
            ("emails", pgettext_lazy("admin ban users form", "E-mails")),
            ("domains", pgettext_lazy("admin ban users form", "E-mail domains")),
        ]

        enable_ip_bans = list(filter(None, [u.joined_from_ip for u in users]))
        if enable_ip_bans:
            self.fields["ban_type"].choices += [
                ("ip", pgettext_lazy("admin ban users form", "IP addresses")),
                (
                    "ip_first",
                    pgettext_lazy(
                        "admin ban users form", "First segment of IP addresses"
                    ),
                ),
                (
                    "ip_two",
                    pgettext_lazy(
                        "admin ban users form", "First two segments of IP addresses"
                    ),
                ),
            ]


class BanForm(forms.ModelForm):
    check_type = forms.TypedChoiceField(
        label=pgettext_lazy("admin ban form", "Check type"),
        coerce=int,
        choices=Ban.CHOICES,
    )
    registration_only = YesNoSwitch(
        label=pgettext_lazy("admin ban form", "Restrict this ban to registrations"),
        help_text=pgettext_lazy(
            "admin ban form",
            "Changing this to yes will make this ban check be only performed on registration step. This is good if you want to block certain registrations like ones from recently compromised e-mail providers, without harming existing users.",
        ),
    )
    banned_value = forms.CharField(
        label=pgettext_lazy("admin ban form", "Banned value"),
        max_length=250,
        help_text=pgettext_lazy(
            "admin ban form",
            'This value is case-insensitive and accepts asterisk (*) for partial matches. For example, making IP ban for value "83.*" will ban all IP addresses beginning with "83.".',
        ),
        error_messages={
            "max_length": pgettext_lazy(
                "admin ban form", "Banned value can't be longer than 250 characters."
            )
        },
    )
    user_message = forms.CharField(
        label=pgettext_lazy("admin ban form", "User message"),
        required=False,
        max_length=1000,
        help_text=pgettext_lazy(
            "admin ban form",
            "Optional message displayed to user instead of default one.",
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": pgettext_lazy(
                "admin ban form", "Message can't be longer than 1000 characters."
            )
        },
    )
    staff_message = forms.CharField(
        label=pgettext_lazy("admin ban form", "Team message"),
        required=False,
        max_length=1000,
        help_text=pgettext_lazy(
            "admin ban form", "Optional ban message for moderators and administrators."
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": pgettext_lazy(
                "admin ban form", "Message can't be longer than 1000 characters."
            )
        },
    )
    expires_on = IsoDateTimeField(
        label=pgettext_lazy("admin ban form", "Expiration date"),
        required=False,
    )

    class Meta:
        model = Ban
        fields = [
            "check_type",
            "registration_only",
            "banned_value",
            "user_message",
            "staff_message",
            "expires_on",
        ]

    def clean_banned_value(self):
        data = self.cleaned_data["banned_value"]
        while "**" in data:
            data = data.replace("**", "*")

        if data == "*":
            raise forms.ValidationError(
                pgettext("admin ban form", "Banned value is too vague.")
            )

        return data


class FilterBansForm(forms.Form):
    check_type = forms.ChoiceField(
        label=pgettext_lazy("admin bans filter form", "Type"),
        required=False,
        choices=[
            ("", pgettext_lazy("admin bans type filter choice", "All bans")),
            ("names", pgettext_lazy("admin bans type filter choice", "Usernames")),
            ("emails", pgettext_lazy("admin bans filter form", "E-mails")),
            ("ips", pgettext_lazy("admin bans type filter choice", "IPs")),
        ],
    )
    value = forms.CharField(
        label=pgettext_lazy("admin bans filter form", "Banned value begins with"),
        required=False,
    )
    registration_only = forms.ChoiceField(
        label=pgettext_lazy("admin bans filter form", "Registration only"),
        required=False,
        choices=[
            ("", pgettext_lazy("admin bans registration filter choice", "Any")),
            ("only", pgettext_lazy("admin bans registration filter choice", "Yes")),
            ("exclude", pgettext_lazy("admin bans registration filter choice", "No")),
        ],
    )
    state = forms.ChoiceField(
        label=pgettext_lazy("admin bans filter form", "State"),
        required=False,
        choices=[
            ("", pgettext_lazy("admin bans state filter choice", "Any")),
            ("used", pgettext_lazy("admin bans state filter choice", "Active")),
            ("unused", pgettext_lazy("admin bans state filter choice", "Expired")),
        ],
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("check_type") == "names":
            queryset = queryset.filter(check_type=0)

        if criteria.get("check_type") == "emails":
            queryset = queryset.filter(check_type=1)

        if criteria.get("check_type") == "ips":
            queryset = queryset.filter(check_type=2)

        if criteria.get("value"):
            queryset = queryset.filter(
                banned_value__startswith=criteria.get("value").lower()
            )

        if criteria.get("state") == "used":
            queryset = queryset.filter(is_checked=True)

        if criteria.get("state") == "unused":
            queryset = queryset.filter(is_checked=False)

        if criteria.get("registration_only") == "only":
            queryset = queryset.filter(registration_only=True)

        if criteria.get("registration_only") == "exclude":
            queryset = queryset.filter(registration_only=False)

        return queryset


class RequestDataDownloadsForm(forms.Form):
    user_identifiers = forms.CharField(
        label=pgettext_lazy("admin data download request form", "Usernames or emails"),
        help_text=pgettext_lazy(
            "admin data download request form",
            "Enter every item in new line. Duplicates will be ignored. This field is case insensitive. Depending on site configuration and amount of data to archive it may take up to few days for requests to complete. E-mail will notification will be sent to every user once their download is ready.",
        ),
        widget=forms.Textarea,
    )

    def clean_user_identifiers(self):
        user_identifiers = self.cleaned_data["user_identifiers"].lower().splitlines()
        user_identifiers = list(filter(bool, user_identifiers))
        user_identifiers = list(set(user_identifiers))

        if len(user_identifiers) > 20:
            raise forms.ValidationError(
                pgettext(
                    "admin data download request form",
                    "You may not enter more than 20 items at a single time (You have entered %(show_value)s).",
                )
                % {"show_value": len(user_identifiers)}
            )

        return user_identifiers

    def clean(self):
        data = super().clean()

        if data.get("user_identifiers"):
            username_match = Q(slug__in=data["user_identifiers"])
            email_match = Q(email_hash__in=map(hash_email, data["user_identifiers"]))

            data["users"] = list(User.objects.filter(username_match | email_match))

            if len(data["users"]) != len(data["user_identifiers"]):
                raise forms.ValidationError(
                    pgettext(
                        "admin data download request form",
                        "One or more specified users could not be found.",
                    )
                )

        return data


class FilterDataDownloadsForm(forms.Form):
    status = forms.ChoiceField(
        label=pgettext_lazy("admin data download requests filter form", "Status"),
        required=False,
        choices=DataDownload.STATUS_CHOICES,
    )
    user = forms.CharField(
        label=pgettext_lazy("admin data download requests filter form", "User"),
        required=False,
    )
    requested_by = forms.CharField(
        label=pgettext_lazy("admin data download requests filter form", "Requested by"),
        required=False,
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("status") is not None:
            queryset = queryset.filter(status=criteria["status"])

        if criteria.get("user"):
            queryset = queryset.filter(user__slug__istartswith=criteria["user"])

        if criteria.get("requested_by"):
            queryset = queryset.filter(
                requester__slug__istartswith=criteria["requested_by"]
            )

        return queryset



from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponsePermanentRedirect, JsonResponse
from django.urls import reverse
from rest_framework.views import exception_handler as rest_exception_handler
from social_core.exceptions import SocialAuthBaseException
from social_core.utils import social_logger

from . import errorpages
from .exceptions import AjaxError, Banned, ExplicitFirstPage, OutdatedSlug

HANDLED_EXCEPTIONS = (
    AjaxError,
    Banned,
    ExplicitFirstPage,
    Http404,
    OutdatedSlug,
    PermissionDenied,
    SocialAuthBaseException,
)


def is_misago_exception(exception):
    return isinstance(exception, HANDLED_EXCEPTIONS)


def handle_ajax_error(request, exception):
    json = {"is_error": 1, "message": str(exception.message)}
    return JsonResponse(json, status=exception.code)


def handle_banned_exception(request, exception):
    return errorpages.banned(request, exception)


def handle_explicit_first_page_exception(request, exception):
    matched_url = request.resolver_match.url_name
    if request.resolver_match.namespace:
        matched_url = "%s:%s" % (request.resolver_match.namespace, matched_url)

    url_kwargs = request.resolver_match.kwargs
    del url_kwargs["page"]

    new_url = reverse(matched_url, kwargs=url_kwargs)
    return HttpResponsePermanentRedirect(new_url)


def handle_http404_exception(request, exception):
    return errorpages.page_not_found(request, exception)


def handle_outdated_slug_exception(request, exception):
    view_name = request.resolver_match.view_name

    model = exception.args[0]
    url_kwargs = request.resolver_match.kwargs
    url_kwargs["slug"] = model.slug

    new_url = reverse(view_name, kwargs=url_kwargs)
    return HttpResponsePermanentRedirect(new_url)


def handle_permission_denied_exception(request, exception):
    return errorpages.permission_denied(request, exception)


def handle_social_auth_exception(request, exception):
    social_logger.error(exception, exc_info=exception)
    return errorpages.social_auth_failed(request, exception)


EXCEPTION_HANDLERS = [
    (AjaxError, handle_ajax_error),
    (Banned, handle_banned_exception),
    (Http404, handle_http404_exception),
    (ExplicitFirstPage, handle_explicit_first_page_exception),
    (OutdatedSlug, handle_outdated_slug_exception),
    (PermissionDenied, handle_permission_denied_exception),
    (SocialAuthBaseException, handle_social_auth_exception),
]


def get_exception_handler(exception):
    for exception_type, handler in EXCEPTION_HANDLERS:
        if isinstance(exception, exception_type):
            return handler
    raise ValueError("%s is not a Misago exception" % exception.__class__.__name__)


def handle_misago_exception(request, exception):
    handler = get_exception_handler(exception)
    return handler(request, exception)


def handle_api_exception(exception, context):
    response = rest_exception_handler(exception, context)
    if response:
        if isinstance(exception, Banned):
            response.data["ban"] = exception.ban.get_serialized_message()
        elif isinstance(exception, PermissionDenied):
            try:
                response.data["detail"] = exception.args[0]
            except IndexError:
                pass
        return response



from django.urls import reverse


def preload_socialauth_json(request):
    request.frontend_context.update(
        {"SOCIAL_AUTH": list_enabled_social_auth_providers(request.socialauth)}
    )

    return {}


def list_enabled_social_auth_providers(socialauth):
    providers = []
    for provider in socialauth.values():
        providers.append(
            {
                "pk": provider["pk"],
                "name": provider["name"],
                "button_text": provider["button_text"],
                "button_color": provider["button_color"],
                "url": reverse(
                    "misago:social-begin", kwargs={"backend": provider["pk"]}
                ),
            }
        )
    return providers



from unittest.mock import Mock

from django.core.exceptions import ValidationError
from django.test import TestCase

from ..validators import validate_post_length, validate_thread_title


class ValidatePostLengthTests(TestCase):
    def test_valid_post_length_passes_validation(self):
        """valid post passes validation"""
        settings = Mock(post_length_min=1, post_length_max=50)
        validate_post_length(settings, "Lorem ipsum dolor met sit amet elit.")

    def test_for_empty_post_validation_error_is_raised(self):
        """empty post is rejected"""
        settings = Mock(post_length_min=3)
        with self.assertRaises(ValidationError):
            validate_post_length(settings, "")

    def test_for_too_short_post_validation_error_is_raised(self):
        """too short post is rejected"""
        settings = Mock(post_length_min=3)
        with self.assertRaises(ValidationError):
            validate_post_length(settings, "a")

    def test_for_too_long_post_validation_error_is_raised(self):
        """too long post is rejected"""
        settings = Mock(post_length_min=1, post_length_max=2)
        with self.assertRaises(ValidationError):
            validate_post_length(settings, "abc")


class ValidateThreadTitleTests(TestCase):
    def test_valid_thread_titles_pass_validation(self):
        """validate_thread_title is ok with valid titles"""
        settings = Mock(thread_title_length_min=1, thread_title_length_max=50)

        VALID_TITLES = ["Lorem ipsum dolor met", "123 456 789 112", "Ugabugagagagagaga"]

        for title in VALID_TITLES:
            validate_thread_title(settings, title)

    def test_for_empty_thread_title_validation_error_is_raised(self):
        """empty title is rejected"""
        settings = Mock(thread_title_length_min=3)
        with self.assertRaises(ValidationError):
            validate_thread_title(settings, "")

    def test_for_too_short_thread_title_validation_error_is_raised(self):
        """too short title is rejected"""
        settings = Mock(thread_title_length_min=3)
        with self.assertRaises(ValidationError):
            validate_thread_title(settings, "a")

    def test_for_too_long_thread_title_validation_error_is_raised(self):
        """too long title is rejected"""
        settings = Mock(thread_title_length_min=1, thread_title_length_max=2)
        with self.assertRaises(ValidationError):
            validate_thread_title(settings, "abc")

    def test_for_unsluggable_thread_title_valdiation_error_is_raised(self):
        """unsluggable title is rejected"""
        settings = Mock(thread_title_length_min=1, thread_title_length_max=9)
        with self.assertRaises(ValidationError):
            validate_thread_title(settings, "-#%^&-")



from .. import test
from ...acl.test import patch_user_acl
from ..models import ThreadParticipant
from .test_privatethreads import PrivateThreadsTestCase


class PrivateThreadViewTests(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(self.category, poster=self.user)
        self.test_link = self.thread.get_absolute_url()

    def test_anonymous(self):
        """anonymous user can't see private thread"""
        self.logout_user()

        response = self.client.get(self.test_link)
        self.assertContains(response, "sign in to use private threads", status_code=403)

    @patch_user_acl({"can_use_private_threads": False})
    def test_no_permission(self):
        """user needs to have permission to see private thread"""
        response = self.client.get(self.test_link)
        self.assertContains(response, "t use private threads", status_code=403)

    def test_no_participant(self):
        """user cant see thread he isn't part of"""
        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_mod_not_reported(self):
        """moderator can't see private thread that has no reports"""
        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 404)

    def test_reported_not_mod(self):
        """non-mod can't see private thread that has reported posts"""
        self.thread.has_reported_posts = True
        self.thread.save()

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 404)

    def test_can_see_owner(self):
        """user can see thread he is owner of"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.client.get(self.test_link)
        self.assertContains(response, self.thread.title)

    def test_can_see_participant(self):
        """user can see thread he is participant of"""
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        response = self.client.get(self.test_link)
        self.assertContains(response, self.thread.title)

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_mod_can_see_reported(self):
        """moderator can see private thread that has reports"""
        self.thread.has_reported_posts = True
        self.thread.save()

        response = self.client.get(self.test_link)
        self.assertContains(response, self.thread.title)



from typing import Optional, Protocol

from django.http import HttpRequest

from ...plugins.hooks import FilterHook
from ...users.models import User


class ValidateUserDataHookAction(Protocol):
    """
    A standard Misago function used for validating the user data, or the next
    filter function from another plugin.

    Should raise a Django's `ValidationError` if data is invalid.

    # Arguments

    ## `request: HttpRequest`

    The request object.

    ## `user: Optional[User]`

    A `User` object associated with `user_data["id"]` or `user_data["email"]`,
    or `None` if it's the user's first time signing in with OAuth and the user's
    account hasn't been created yet.

    ## `user_data: dict`

    A Python `dict` with user data extracted from the OAuth 2 server's response:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```

    This `dict` will be unfiltered unless it was filtered by an `action` call or `filter_user_data` was used by the plugin to filter it.

    ## `response_json: dict`

    A Python `dict` with the unfiltered OAuth 2 server's user info JSON.

    # Return value

    A Python `dict` containing validated user data:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```
    """

    def __call__(
        self,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
        response_json: dict,
    ) -> dict:
        pass


class ValidateUserDataHookFilter(Protocol):
    """
    A function implemented by a plugin that can be registered in this hook.

    Should raise a Django's `ValidationError` if data is invalid.

    # Arguments

    ## `action: ValidateUserDataHookAction`

    A standard Misago function used for filtering the user data, or the next
    filter function from another plugin.

    See the [action](#action) section for details.

    ## `request: HttpRequest`

    The request object.

    ## `user: Optional[User]`

    A `User` object associated with `user_data["id"]` or `user_data["email"]`,
    or `None` if it's the user's first time signing in with OAuth and the user's
    account hasn't been created yet.

    ## `user_data: dict`

    A Python `dict` with user data extracted from the OAuth 2 server's response:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```

    This `dict` will be unfiltered unless it was filtered by an `action` call or `filter_user_data` was used by the plugin to filter it.

    ## `response_json: dict`

    A Python `dict` with the unfiltered OAuth 2 server's user info JSON.

    # Return value

    A Python `dict` containing validated user data:

    ```python
    class UserData(TypedDict):
        id: str
        name: str | None
        email: str | None
        avatar: str | None
    ```
    """

    def __call__(
        self,
        action: ValidateUserDataHookAction,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
        response_json: dict,
    ) -> dict:
        pass


class ValidateUserDataHook(
    FilterHook[ValidateUserDataHookAction, ValidateUserDataHookFilter]
):
    """
    This hook wraps the standard function that Misago uses to validate a Python
    `dict` containing the user data extracted from the OAuth 2 server's response.

    Should raise a Django's `ValidationError` if data is invalid.

    # Example

    The code below implements a custom validator function that extends the standard
    logic with additional check for a permission to use the forum by the user:

    ```python
    @validate_user_data_hook.append_filter
    def normalize_gmail_email(
        action,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
        response_json: dict,
    ) -> dict:
        # Prevent user from completing the OAuth 2 flow unless they are a member
        # of the "forum" group
        if (
            not response_json.get("groups")
            or not isinstance(response_json["groups"], list)
            or not "forum" in response_json["groups"]
        ):
            raise ValidationError("You don't have a permission to use the forums.")

        # Call the next function in chain
        return action(user_data, request, user, user_data, response_json)
    ```
    """

    __slots__ = FilterHook.__slots__

    def __call__(
        self,
        action: ValidateUserDataHookAction,
        request: HttpRequest,
        user: Optional[User],
        user_data: dict,
        response_json: dict,
    ) -> dict:
        return super().__call__(action, request, user, user_data, response_json)


validate_user_data_hook = ValidateUserDataHook()



import random
import time

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from faker import Factory

from ....categories.models import Category
from ....core.management.progressbar import show_progress
from ....core.pgutils import chunk_queryset
from ....threads.models import Thread
from ...posts import get_fake_hidden_post, get_fake_post, get_fake_unapproved_post

User = get_user_model()


class Command(BaseCommand):
    help = "Creates random posts for dev and testing purposes."

    def add_arguments(self, parser):
        parser.add_argument(
            "posts", help="number of posts to create", nargs="?", type=int, default=5
        )

    def handle(self, *args, **options):
        items_to_create = options["posts"]
        fake = Factory.create()

        message = "Creating %s fake posts...\n"
        self.stdout.write(message % items_to_create)

        created_posts = 0
        start_time = time.time()
        show_progress(self, created_posts, items_to_create)

        while created_posts < items_to_create:
            thread = Thread.objects.order_by("?")[:1].first()

            # 10% chance poster is anonymous
            if random.randint(0, 100) > 90:
                poster = None
            else:
                poster = User.objects.order_by("?").last()

            # There's 5% chance post is unapproved
            if random.randint(0, 100) > 90:
                get_fake_unapproved_post(fake, thread, poster)

            # There's further 5% chance post is hidden
            elif random.randint(0, 100) > 95:
                if random.randint(0, 100) > 90:
                    hidden_by = None
                else:
                    hidden_by = User.objects.order_by("?").last()

                get_fake_hidden_post(fake, thread, poster, hidden_by)

            # Default, standard post
            else:
                get_fake_post(fake, thread, poster)

            created_posts += 1
            show_progress(self, created_posts, items_to_create, start_time)

        for thread in chunk_queryset(Thread.objects.all()):
            thread.synchronize()
            thread.save()

        for category in Category.objects.all():
            category.synchronize()
            category.save()

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\nSuccessfully created %s fake posts in %s"
        self.stdout.write(message % (created_posts, total_humanized))



from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext
from rest_framework.response import Response

from ...serializers import MovePostsSerializer


def posts_move_endpoint(request, thread, viewmodel):
    if not thread.acl["can_move_posts"]:
        raise PermissionDenied(
            pgettext("posts api", "You can't move posts from this thread.")
        )

    serializer = MovePostsSerializer(
        data=request.data,
        context={
            "request": request,
            "settings": request.settings,
            "thread": thread,
            "viewmodel": viewmodel,
        },
    )

    if not serializer.is_valid():
        if "new_thread" in serializer.errors:
            errors = serializer.errors["new_thread"]
        else:
            errors = list(serializer.errors.values())[0]
        # Fix for KeyError - errors[0]
        try:
            return Response({"detail": errors[0]}, status=400)
        except KeyError:
            return Response({"detail": list(errors.values())[0][0]}, status=400)

    new_thread = serializer.validated_data["new_thread"]

    for post in serializer.validated_data["posts"]:
        post.move(new_thread)
        post.save()

    thread.synchronize()
    thread.save()

    new_thread.synchronize()
    new_thread.save()

    thread.category.synchronize()
    thread.category.save()

    if thread.category != new_thread.category:
        new_thread.category.synchronize()
        new_thread.category.save()

    return Response({})



import json
from hashlib import sha256

from django.templatetags.static import static
from django.urls import reverse
from django.utils.translation import get_language

import misago
from . import settings


# Simple but hard to guess version signature for current Misago version
# Used to cache-bust django-i18n.js URL in the browser
I18N_VERSION_SIGNATURE = sha256(
    (
        f"{misago.__version__}{misago.__released__}"
        f"{settings.LANGUAGE_CODE}{settings.SECRET_KEY}"
    ).encode()
).hexdigest()


def conf(request):
    return {
        "BLANK_AVATAR_URL": (
            request.settings.blank_avatar or static(settings.MISAGO_BLANK_AVATAR)
        ),
        "DEBUG": settings.DEBUG,
        "I18N_VERSION_SIGNATURE": I18N_VERSION_SIGNATURE,
        "LANGUAGE_CODE_SHORT": get_language()[:2],
        "LOGIN_REDIRECT_URL": settings.LOGIN_REDIRECT_URL,
        "LOGIN_URL": settings.LOGIN_URL,
        "LOGOUT_URL": settings.LOGOUT_URL,
        "THREADS_ON_INDEX": settings.MISAGO_THREADS_ON_INDEX,
        "CSRF_COOKIE_NAME": json.dumps(settings.CSRF_COOKIE_NAME),
        "settings": request.settings,
    }


def og_image(request):
    og_image = request.settings.get("og_image")
    if not og_image["value"]:
        return {"og_image": None}

    return {
        "og_image": {
            "url": og_image["value"],
            "width": og_image["width"],
            "height": og_image["height"],
        }
    }


def preload_settings_json(request):
    preloaded_settings = request.settings.get_public_settings()

    delegate_auth = request.settings.enable_oauth2_client

    if request.settings.enable_oauth2_client:
        login_url = reverse("misago:oauth2-login")
    else:
        login_url = reverse(settings.LOGIN_URL)

    preloaded_settings.update(
        {
            "DELEGATE_AUTH": delegate_auth,
            "LOGIN_API_URL": settings.MISAGO_LOGIN_API_URL,
            "LOGIN_REDIRECT_URL": reverse(settings.LOGIN_REDIRECT_URL),
            "LOGIN_URL": login_url,
            "LOGOUT_URL": reverse(settings.LOGOUT_URL),
        }
    )

    request.frontend_context.update(
        {
            "BLANK_AVATAR_URL": (
                request.settings.blank_avatar or static(settings.MISAGO_BLANK_AVATAR)
            ),
            "CSRF_COOKIE_NAME": settings.CSRF_COOKIE_NAME,
            "ENABLE_DELETE_OWN_ACCOUNT": (
                not delegate_auth and request.settings.allow_delete_own_account
            ),
            "ENABLE_DOWNLOAD_OWN_DATA": request.settings.allow_data_downloads,
            "MISAGO_PATH": reverse("misago:index"),
            "SETTINGS": preloaded_settings,
            "STATIC_URL": settings.STATIC_URL,
            "THREADS_ON_INDEX": settings.MISAGO_THREADS_ON_INDEX,
            "NOTIFICATIONS_API": reverse("misago:apiv2:notifications"),
            "NOTIFICATIONS_URL": reverse("misago:notifications"),
        }
    )

    if (
        request.user.is_authenticated
        and request.user.is_staff
        and request.settings.show_admin_panel_link_in_ui
    ):
        request.frontend_context["ADMIN_URL"] = reverse("misago:admin:index")

    return {}



# pylint: disable=keyword-arg-before-vararg
import json

from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.http import JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import pgettext
from social_core.pipeline.partial import partial
from unidecode import unidecode

from ..core.exceptions import SocialAuthBanned, SocialAuthFailed
from ..legal.models import Agreement
from ..users.bans import get_request_ip_ban, get_user_ban
from ..users.forms.register import SocialAuthRegisterForm
from ..users.models import Ban
from ..users.registration import (
    get_registration_result_json,
    save_user_agreements,
    send_welcome_email,
)
from ..users.setupnewuser import setup_new_user
from ..users.validators import (
    ValidationError,
    validate_email,
    validate_new_registration,
    validate_username,
)
from .providers import providers

User = get_user_model()


def validate_ip_not_banned(strategy, details, backend, user=None, *args, **kwargs):
    """
    Pipeline step that interrupts pipeline if found user is non-staff and IP banned
    """
    if not user or user.is_staff:
        return None

    ban = get_request_ip_ban(strategy.request)
    if ban:
        hydrated_ban = Ban(
            check_type=Ban.IP, user_message=ban["message"], expires_on=ban["expires_on"]
        )
        raise SocialAuthBanned(backend, hydrated_ban)


def validate_user_not_banned(strategy, details, backend, user=None, *args, **kwargs):
    """Pipeline step that interrupts pipeline if found user is non-staff and banned"""
    if not user or user.is_staff:
        return None

    user_ban = get_user_ban(user, strategy.request.cache_versions)
    if user_ban:
        raise SocialAuthBanned(backend, user_ban)


def perpare_username(username):
    return "".join(filter(str.isalnum, unidecode(username)))


def associate_by_email(strategy, details, backend, user=None, *args, **kwargs):
    """If user with e-mail from provider exists in database and is active,
    this step authenticates them.
    """
    enable_step = strategy.setting("ASSOCIATE_BY_EMAIL", default=False, backend=backend)

    if user or not enable_step:
        return None

    email = details.get("email")
    if not email:
        return None

    try:
        user = User.objects.get_by_email(email)
    except User.DoesNotExist:
        return None

    backend_name = providers.get_name(backend.name)

    if not user.is_active:
        raise SocialAuthFailed(
            backend,
            pgettext(
                "social auth",
                "The e-mail address associated with your %(backend)s account is not available for use on this site.",
            )
            % {"backend": backend_name},
        )

    if user.requires_activation_by_admin:
        raise SocialAuthFailed(
            backend,
            pgettext(
                "social auth",
                "Your account has to be activated by site administrator before you will be able to sign in with %(backend)s.",
            )
            % {"backend": backend_name},
        )

    return {"user": user, "is_new": False}


def get_username(strategy, details, backend, user=None, *args, **kwargs):
    """Resolve valid username for use in new account"""
    if user:
        return None

    settings = strategy.request.settings

    username = perpare_username(details.get("username", ""))
    full_name = perpare_username(details.get("full_name", ""))
    first_name = perpare_username(details.get("first_name", ""))
    last_name = perpare_username(details.get("last_name", ""))

    names_to_try = [username, first_name]

    if username:
        names_to_try.append(username)

    if first_name:
        names_to_try.append(first_name)
        if last_name:
            # if first name is taken, try first name + first char of last name
            names_to_try.append(first_name + last_name[0])

    if full_name:
        names_to_try.append(full_name)

    username_length_max = settings.username_length_max
    for name in names_to_try:
        if len(name) > username_length_max:
            names_to_try.append(name[:username_length_max])

    for name in filter(bool, names_to_try):
        try:
            validate_username(settings, name)
            return {"clean_username": name}
        except ValidationError:
            pass


def create_user(strategy, details, backend, user=None, *args, **kwargs):
    """Aggressively attempt to register and sign in new user"""
    if user:
        return None

    request = strategy.request
    settings = request.settings

    email = details.get("email")
    username = kwargs.get("clean_username")

    if not email or not username:
        return None

    try:
        validate_email(email)
        validate_new_registration(request, {"email": email, "username": username})
    except ValidationError:
        return None

    activation_kwargs = {}
    if settings.account_activation == "admin":
        activation_kwargs = {"requires_activation": User.ACTIVATION_ADMIN}

    new_user = User.objects.create_user(
        username, email, joined_from_ip=request.user_ip, **activation_kwargs
    )

    setup_new_user(settings, new_user)
    send_welcome_email(request, new_user)

    return {"user": new_user, "is_new": True}


@partial
def create_user_with_form(strategy, details, backend, user=None, *args, **kwargs):
    """
    create_user lets user confirm account creation by entering final username or email
    """
    if user:
        return None

    request = strategy.request
    settings = request.settings
    backend_name = providers.get_name(backend.name)

    if request.method == "POST":
        try:
            request_data = json.loads(request.body)
        except (TypeError, ValueError):
            request_data = request.POST.copy()

        form = SocialAuthRegisterForm(
            request_data, request=request, agreements=Agreement.objects.get_agreements()
        )

        if not form.is_valid():
            return JsonResponse(form.errors, status=400)

        email_verified = form.cleaned_data["email"] == details.get("email")

        activation_kwargs = {}
        if settings.account_activation == "admin":
            activation_kwargs = {"requires_activation": User.ACTIVATION_ADMIN}
        elif settings.account_activation == "user" and not email_verified:
            activation_kwargs = {"requires_activation": User.ACTIVATION_USER}

        try:
            new_user = User.objects.create_user(
                form.cleaned_data["username"],
                form.cleaned_data["email"],
                joined_from_ip=request.user_ip,
                **activation_kwargs
            )
            setup_new_user(settings, new_user)
        except IntegrityError:
            return JsonResponse(
                {
                    "__all__": pgettext(
                        "social auth", "Please try resubmitting the form."
                    )
                },
                status=400,
            )

        save_user_agreements(new_user, form)
        send_welcome_email(request, new_user)

        return {"user": new_user, "is_new": True}

    request.frontend_context["SOCIAL_AUTH_FORM"] = {
        "backend_name": backend_name,
        "step": "register",
        "email": details.get("email"),
        "username": kwargs.get("clean_username"),
        "url": reverse("misago:social-complete", kwargs={"backend": backend.name}),
    }

    return render(request, "misago/socialauth.html", {"backend_name": backend_name})


@partial
def require_activation(
    strategy, details, backend, user=None, is_new=False, *args, **kwargs
):
    if not user:
        # Social auth pipeline has entered corrupted state
        # Remove partial auth state and redirect user to beginning
        partial_token = strategy.session.get("partial_pipeline_token")
        if partial_token:
            strategy.clean_partial_pipeline(partial_token)
        return None

    if not user.requires_activation:
        return None

    request = strategy.request
    backend_name = providers.get_name(backend.name)

    response_data = get_registration_result_json(user)
    response_data.update({"step": "done", "backend_name": backend_name})

    if request.method == "POST":
        # we are carrying on from requestration request
        return JsonResponse(response_data)

    request.frontend_context["SOCIAL_AUTH_FORM"] = response_data
    request.frontend_context["SOCIAL_AUTH_FORM"].update(
        {"url": reverse("misago:social-complete", kwargs={"backend": backend.name})}
    )

    return render(request, "misago/socialauth.html", {"backend_name": backend_name})



from ..models import Notification
from ..users import notify_user


def test_notify_user_creates_notification_for_user(user):
    notification = notify_user(user, "TEST")
    assert notification
    assert notification.user == user
    assert notification.verb == "TEST"
    assert notification.actor is None
    assert notification.actor_name is None
    assert notification.category is None
    assert notification.thread is None
    assert notification.thread_title is None
    assert notification.post is None

    db_notification = Notification.objects.get(id=notification.id)
    assert db_notification.user == user
    assert db_notification.verb == "TEST"
    assert db_notification.actor is None
    assert db_notification.actor_name is None
    assert db_notification.category is None
    assert db_notification.thread is None
    assert db_notification.thread_title is None
    assert db_notification.post is None


def test_notify_user_increases_user_unread_notifications_counter(user):
    assert user.unread_notifications == 0
    assert notify_user(user, "TEST")

    user.refresh_from_db()
    assert user.unread_notifications == 1


def test_notify_user_sets_actor_on_notification(user, other_user):
    notification = notify_user(user, "TEST", actor=other_user)
    assert notification
    assert notification.user == user
    assert notification.verb == "TEST"
    assert notification.actor == other_user
    assert notification.actor_name == other_user.username
    assert notification.category is None
    assert notification.thread is None
    assert notification.thread_title is None
    assert notification.post is None

    db_notification = Notification.objects.get(id=notification.id)
    assert db_notification.user == user
    assert db_notification.verb == "TEST"
    assert db_notification.actor == other_user
    assert db_notification.actor_name == other_user.username
    assert db_notification.category is None
    assert db_notification.thread is None
    assert db_notification.thread_title is None
    assert db_notification.post is None


def test_notify_user_sets_category_on_notification(user, default_category):
    notification = notify_user(user, "TEST", category=default_category)
    assert notification
    assert notification.user == user
    assert notification.verb == "TEST"
    assert notification.actor is None
    assert notification.actor_name is None
    assert notification.category == default_category
    assert notification.thread is None
    assert notification.thread_title is None
    assert notification.post is None

    db_notification = Notification.objects.get(id=notification.id)
    assert db_notification.user == user
    assert db_notification.verb == "TEST"
    assert db_notification.actor is None
    assert db_notification.actor_name is None
    assert db_notification.category == default_category
    assert db_notification.thread is None
    assert db_notification.thread_title is None
    assert db_notification.post is None


def test_notify_user_sets_thread_on_notification(user, thread):
    notification = notify_user(user, "TEST", thread=thread)
    assert notification
    assert notification.user == user
    assert notification.verb == "TEST"
    assert notification.actor is None
    assert notification.actor_name is None
    assert notification.category is None
    assert notification.thread == thread
    assert notification.thread_title == thread.title
    assert notification.post is None

    db_notification = Notification.objects.get(id=notification.id)
    assert db_notification.user == user
    assert db_notification.verb == "TEST"
    assert db_notification.actor is None
    assert db_notification.actor_name is None
    assert db_notification.category is None
    assert db_notification.thread == thread
    assert db_notification.thread_title == thread.title
    assert db_notification.post is None


def test_notify_user_sets_post_on_notification(user, post):
    notification = notify_user(user, "TEST", post=post)
    assert notification
    assert notification.user == user
    assert notification.verb == "TEST"
    assert notification.actor is None
    assert notification.actor_name is None
    assert notification.category is None
    assert notification.thread is None
    assert notification.thread_title is None
    assert notification.post == post

    db_notification = Notification.objects.get(id=notification.id)
    assert db_notification.user == user
    assert db_notification.verb == "TEST"
    assert db_notification.actor is None
    assert db_notification.actor_name is None
    assert db_notification.category is None
    assert db_notification.thread is None
    assert db_notification.thread_title is None
    assert db_notification.post == post



from django.contrib.auth import get_user_model
from django.shortcuts import get_object_or_404
from rest_framework.response import Response

from ....core.shortcuts import get_int_or_404
from ...models import Rank
from ...serializers import UserCardSerializer
from ...viewmodels import ActivePosters, RankUsers

User = get_user_model()


def active(request):
    users = ActivePosters(request)
    return Response(users.get_frontend_context())


def rank_users(request):
    rank_pk = get_int_or_404(request.query_params.get("rank"))
    rank = get_object_or_404(Rank.objects, pk=rank_pk, is_tab=True)

    page = get_int_or_404(request.GET.get("page", 0))
    if page == 1:
        page = 0  # api allows explicit first page

    users = RankUsers(request, rank, page)
    return Response(users.get_frontend_context())


LISTS = {"active": active}


def list_endpoint(request):
    list_type = request.query_params.get("list")
    list_handler = LISTS.get(list_type)

    if list_handler:
        return list_handler(request)
    return rank_users(request)


ScoredUserSerializer = UserCardSerializer.extend_fields("meta")



from django.utils.module_loading import import_string

from .. import hooks
from ..conf import settings

filters_list = settings.MISAGO_POST_SEARCH_FILTERS
SEARCH_FILTERS = list(map(import_string, filters_list))


def filter_search(search, filters=None):
    filters = filters or SEARCH_FILTERS

    for search_filter in filters:
        search = search_filter(search) or search
    for search_filter in hooks.post_search_filters:
        search = search_filter(search) or search

    return search



from django.db import models

from .providers import providers


class SocialAuthProvider(models.Model):
    provider = models.CharField(primary_key=True, max_length=30)
    button_text = models.CharField(max_length=255, null=True, blank=True)
    button_color = models.CharField(max_length=7, null=True, blank=True)
    settings = models.JSONField(default=dict)
    is_active = models.BooleanField(default=False)
    order = models.IntegerField(default=0)

    class Meta:
        ordering = ["order"]

    def __str__(self):
        return providers.get_name(self.provider)



import warnings

from django.test import TestCase

from .. import SUPPORTED_ENGINES, check_db_engine

INVALID_ENGINES = [
    "django.db.backends.sqlite3",
    "django.db.backends.mysql",
    "django.db.backends.oracle",
]


class TestCheckDBEngine(TestCase):
    def test_valid_engines(self):
        """check passes valid engines"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            for engine in SUPPORTED_ENGINES:
                with self.settings(DATABASES={"default": {"ENGINE": engine}}):
                    errors = check_db_engine(None)
                    self.assertEqual(errors, [])

    def test_invalid_engines(self):
        """check returns error for invalid engines"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            for engine in INVALID_ENGINES:
                with self.settings(DATABASES={"default": {"ENGINE": engine}}):
                    errors = check_db_engine(None)
                    self.assertTrue(errors)



# Generated by Django 4.2.7 on 2023-11-13 19:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("misago_categories", "0009_auto_20221101_2111"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rolecategoryacl",
            name="category_role",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="misago_categories.categoryrole",
            ),
        ),
    ]



from django.urls import reverse

from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..test import patch_category_acl


class StartThreadTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.api_link = reverse("misago:api:thread-list")

    def test_cant_start_thread_as_guest(self):
        """user has to be authenticated to be able to post thread"""
        self.logout_user()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)

    @patch_category_acl({"can_see": False})
    def test_cant_see(self):
        """has no permission to see selected category"""
        response = self.client.post(self.api_link, {"category": self.category.pk})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "category": ["Selected category is invalid."],
                "post": ["You have to enter a message."],
                "title": ["You have to enter a thread title."],
            },
        )

    @patch_category_acl({"can_browse": False})
    def test_cant_browse(self):
        """has no permission to browse selected category"""
        response = self.client.post(self.api_link, {"category": self.category.pk})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "category": ["Selected category is invalid."],
                "post": ["You have to enter a message."],
                "title": ["You have to enter a thread title."],
            },
        )

    @patch_category_acl({"can_start_threads": False})
    def test_cant_start_thread(self):
        """permission to start thread in category is validated"""
        response = self.client.post(self.api_link, {"category": self.category.pk})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "category": [
                    "You don't have permission to start new threads in this category."
                ],
                "post": ["You have to enter a message."],
                "title": ["You have to enter a thread title."],
            },
        )

    @patch_category_acl({"can_start_threads": True, "can_close_threads": False})
    def test_cant_start_thread_in_locked_category(self):
        """can't post in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.post(self.api_link, {"category": self.category.pk})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "category": [
                    "This category is closed. You can't start new threads in it."
                ],
                "post": ["You have to enter a message."],
                "title": ["You have to enter a thread title."],
            },
        )

    def test_cant_start_thread_in_invalid_category(self):
        """can't post in invalid category"""
        response = self.client.post(
            self.api_link, {"category": self.category.pk * 100000}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "category": [
                    "Selected category doesn't exist or "
                    "you don't have permission to browse it."
                ],
                "post": ["You have to enter a message."],
                "title": ["You have to enter a thread title."],
            },
        )

    @patch_category_acl({"can_start_threads": True})
    def test_empty_data(self):
        """no data sent handling has no showstoppers"""
        response = self.client.post(self.api_link, data={})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "category": ["You have to select category to post thread in."],
                "title": ["You have to enter a thread title."],
                "post": ["You have to enter a message."],
            },
        )

    @patch_category_acl({"can_start_threads": True})
    def test_invalid_data(self):
        """api errors for invalid request data"""
        response = self.client.post(
            self.api_link, "false", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    @patch_category_acl({"can_start_threads": True})
    def test_title_is_validated(self):
        """title is validated"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "------",
                "post": "Lorem ipsum dolor met, sit amet elit!",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"title": ["Thread title should contain alpha-numeric characters."]},
        )

    @patch_category_acl({"can_start_threads": True})
    def test_post_is_validated(self):
        """post is validated"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Lorem ipsum dolor met",
                "post": "a",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "post": [
                    "Posted message should be at least 5 characters long (it has 1)."
                ]
            },
        )

    @patch_category_acl({"can_start_threads": True})
    def test_can_start_thread(self):
        """endpoint creates new thread"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]

        response_json = response.json()
        self.assertEqual(response_json["url"], thread.get_absolute_url())

        response = self.client.get(thread.get_absolute_url())
        self.assertContains(response, self.category.name)
        self.assertContains(response, thread.title)
        self.assertContains(response, "<p>Lorem ipsum dolor met!</p>")

        # api increased user's threads and posts counts
        self.reload_user()
        self.assertEqual(self.user.threads, 1)
        self.assertEqual(self.user.posts, 1)

        self.assertEqual(self.user.audittrail_set.count(), 1)

        self.assertEqual(thread.category_id, self.category.pk)
        self.assertEqual(thread.title, "Hello, I am test thread!")
        self.assertEqual(thread.starter_id, self.user.id)
        self.assertEqual(thread.starter_name, self.user.username)
        self.assertEqual(thread.starter_slug, self.user.slug)
        self.assertEqual(thread.last_poster_id, self.user.id)
        self.assertEqual(thread.last_poster_name, self.user.username)
        self.assertEqual(thread.last_poster_slug, self.user.slug)

        post = self.user.post_set.all()[:1][0]
        self.assertEqual(post.category_id, self.category.pk)
        self.assertEqual(post.original, "Lorem ipsum dolor met!")
        self.assertEqual(post.poster_id, self.user.id)
        self.assertEqual(post.poster_name, self.user.username)

        category = Category.objects.get(pk=self.category.pk)
        self.assertEqual(category.threads, 1)
        self.assertEqual(category.posts, 1)
        self.assertEqual(category.last_thread_id, thread.id)
        self.assertEqual(category.last_thread_title, thread.title)
        self.assertEqual(category.last_thread_slug, thread.slug)

        self.assertEqual(category.last_poster_id, self.user.id)
        self.assertEqual(category.last_poster_name, self.user.username)
        self.assertEqual(category.last_poster_slug, self.user.slug)

    @patch_category_acl({"can_start_threads": True, "can_close_threads": False})
    def test_start_closed_thread_no_permission(self):
        """permission is checked before thread is closed"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "close": True,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertFalse(thread.is_closed)

    @patch_category_acl({"can_start_threads": True, "can_close_threads": True})
    def test_start_closed_thread(self):
        """can post closed thread"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "close": True,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertTrue(thread.is_closed)

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 1})
    def test_start_unpinned_thread(self):
        """can post unpinned thread"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "pin": 0,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertEqual(thread.weight, 0)

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 1})
    def test_start_locally_pinned_thread(self):
        """can post locally pinned thread"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "pin": 1,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertEqual(thread.weight, 1)

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 2})
    def test_start_globally_pinned_thread(self):
        """can post globally pinned thread"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "pin": 2,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertEqual(thread.weight, 2)

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 1})
    def test_start_globally_pinned_thread_no_permission(self):
        """cant post globally pinned thread without permission"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "pin": 2,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertEqual(thread.weight, 0)

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 0})
    def test_start_locally_pinned_thread_no_permission(self):
        """cant post locally pinned thread without permission"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "pin": 1,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertEqual(thread.weight, 0)

    @patch_category_acl({"can_start_threads": True, "can_hide_threads": 1})
    def test_start_hidden_thread(self):
        """can post hidden thread"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "hide": 1,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertTrue(thread.is_hidden)

        category = Category.objects.get(pk=self.category.pk)
        self.assertNotEqual(category.last_thread_id, thread.id)

    @patch_category_acl({"can_start_threads": True, "can_hide_threads": 0})
    def test_start_hidden_thread_no_permission(self):
        """cant post hidden thread without permission"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
                "hide": 1,
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertFalse(thread.is_hidden)

    @patch_category_acl({"can_start_threads": True})
    def test_post_unicode(self):
        """unicode characters can be posted"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Brzęczyżczykiewicz",
                "post": "Chrzążczyżewoszyce, powiat Łękółody.",
            },
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_start_threads": True})
    def test_category_moderation_queue(self):
        """start unapproved thread in category that requires approval"""
        self.category.require_threads_approval = True
        self.category.save()

        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertTrue(thread.is_unapproved)
        self.assertTrue(thread.has_unapproved_posts)

        post = self.user.post_set.all()[:1][0]
        self.assertTrue(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts)
        self.assertFalse(category.last_thread_id == thread.id)

    @patch_category_acl({"can_start_threads": True})
    @patch_user_acl({"can_approve_content": True})
    def test_category_moderation_queue_bypass(self):
        """bypass moderation queue due to user's acl"""
        self.category.require_threads_approval = True
        self.category.save()

        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads + 1)
        self.assertEqual(category.posts, self.category.posts + 1)
        self.assertEqual(category.last_thread_id, thread.id)

    @patch_category_acl({"can_start_threads": True, "require_threads_approval": True})
    def test_user_moderation_queue(self):
        """start unapproved thread in category that requires approval"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertTrue(thread.is_unapproved)
        self.assertTrue(thread.has_unapproved_posts)

        post = self.user.post_set.all()[:1][0]
        self.assertTrue(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts)
        self.assertFalse(category.last_thread_id == thread.id)

    @patch_category_acl({"can_start_threads": True, "require_threads_approval": True})
    @patch_user_acl({"can_approve_content": True})
    def test_user_moderation_queue_bypass(self):
        """bypass moderation queue due to user's acl"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads + 1)
        self.assertEqual(category.posts, self.category.posts + 1)
        self.assertEqual(category.last_thread_id, thread.id)

    @patch_category_acl(
        {
            "can_start_threads": True,
            "require_replies_approval": True,
            "require_edits_approval": True,
        }
    )
    def test_omit_other_moderation_queues(self):
        """other queues are omitted"""
        self.category.require_replies_approval = True
        self.category.require_edits_approval = True
        self.category.save()

        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads + 1)
        self.assertEqual(category.posts, self.category.posts + 1)
        self.assertEqual(category.last_thread_id, thread.id)



import pytest

from ....cache.test import assert_invalidates_cache
from ... import THEME_CACHE
from ..css import change_css_source, get_theme_media_map, rebuild_css
from ..tasks import build_single_theme_css, build_theme_css


@pytest.fixture
def assert_snapshot_match(snapshot, theme):
    def _assert_snapshot_match(result):
        result = result.replace(theme.dirname, "themedir")
        assert snapshot == result

    return _assert_snapshot_match


@pytest.fixture
def media_map(theme, image):
    return get_theme_media_map(theme)


def test_tasks_builds_single_css_file(theme, image, css_needing_build):
    build_single_theme_css(css_needing_build.pk)
    css_needing_build.refresh_from_db()
    assert css_needing_build.build_file


def test_tasks_skips_single_css_file_that_doesnt_require_build(theme, css):
    build_single_theme_css(css.pk)
    css.refresh_from_db()
    assert not css.build_file


def test_tasks_handles_nonexisting_css_file(db):
    build_single_theme_css(1)


def test_tasks_builds_theme_css_files_that_require_it(theme, image, css_needing_build):
    build_theme_css(theme.pk)
    css_needing_build.refresh_from_db()
    assert css_needing_build.build_file


def test_tasks_skips_theme_css_files_that_dont_require_build(theme, css):
    build_theme_css(theme.pk)
    css.refresh_from_db()
    assert not css.build_file


def test_tasks_handles_nonexisting_theme(nonexisting_theme):
    build_theme_css(nonexisting_theme.pk)


def test_media_map_for_theme_without_any_media_files_returns_empty_dict(theme):
    assert get_theme_media_map(theme) == {}


def test_media_map_for_theme_with_media_files_returns_dict_with_data(
    theme, image, media
):
    assert get_theme_media_map(theme)


def test_css_file_is_build(media_map, css_needing_build):
    rebuild_css(media_map, css_needing_build)
    css_needing_build.refresh_from_db()
    assert css_needing_build.build_file


def test_build_css_file_is_hashed(media_map, css_needing_build):
    rebuild_css(media_map, css_needing_build)
    css_needing_build.refresh_from_db()
    assert css_needing_build.build_hash


def test_build_css_file_includes_hash_in_filename(media_map, css_needing_build):
    rebuild_css(media_map, css_needing_build)
    css_needing_build.refresh_from_db()
    assert css_needing_build.build_hash in str(css_needing_build.build_file)


def test_build_css_file_has_size_set(media_map, css_needing_build):
    rebuild_css(media_map, css_needing_build)
    css_needing_build.refresh_from_db()
    assert css_needing_build.size


def test_simple_url_to_file_is_replaced_with_valid_url(
    assert_snapshot_match, media_map, image
):
    css = ".page-header { background-image: url(%s); }" % image.name
    result = change_css_source(media_map, css)
    assert_snapshot_match(result)


def test_relative_url_to_file_is_replaced_with_valid_url(
    assert_snapshot_match, media_map, image
):
    css = ".page-header { background-image: url(./%s); }" % image.name
    result = change_css_source(media_map, css)
    assert_snapshot_match(result)


def test_url_to_file_from_create_react_app_is_replaced_with_valid_url(
    assert_snapshot_match, media_map, image
):
    hashed_name = str(image.file).split("/")[-1]
    css = ".page-header { background-image: url(/static/media/%s); }" % hashed_name
    result = change_css_source(media_map, css)
    assert_snapshot_match(result)


def test_quoted_url_to_file_is_replaced_with_valid_url(
    assert_snapshot_match, media_map, image
):
    css = '.page-header { background-image: url("%s"); }' % image.name
    result = change_css_source(media_map, css)
    assert_snapshot_match(result)


def test_single_quoted_url_to_file_is_replaced_with_valid_url(
    assert_snapshot_match, media_map, image
):
    css = ".page-header { background-image: url('%s'); }" % image.name
    result = change_css_source(media_map, css)
    assert_snapshot_match(result)


def test_absolute_https_url_to_file_is_not_replaced(media_map):
    css = ".page-header { background-image: url(https://cdn.example.com/bg.png); }"
    result = change_css_source(media_map, css)
    assert result == css


def test_absolute_http_url_to_file_is_not_replaced(media_map):
    css = ".page-header { background-image: url(http://cdn.example.com/bg.png); }"
    result = change_css_source(media_map, css)
    assert result == css


def test_absolute_protocol_relative_url_to_file_is_not_replaced(media_map):
    css = ".page-header { background-image: url(://cdn.example.com/bg.png); }"
    result = change_css_source(media_map, css)
    assert result == css


def test_css_file_with_multiple_different_urls_is_correctly_replaced(
    assert_snapshot_match, media_map, image
):
    css = (
        ".page-header { background-image: url(http://cdn.example.com/bg.png); }"
        '\n.container { background-image: url("%s"); }'
        '\n.alert { background-image: url("%s"); }'
    ) % (image.name, str(image.file).strip("/")[-1])

    result = change_css_source(media_map, css)
    assert_snapshot_match(result)


def test_building_single_theme_css_invalidates_theme_cache(
    theme, image, css_needing_build
):
    with assert_invalidates_cache(THEME_CACHE):
        build_single_theme_css(css_needing_build.pk)


def test_building_theme_css_invalidates_theme_cache(theme):
    with assert_invalidates_cache(THEME_CACHE):
        build_theme_css(theme.pk)



import time

from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress
from ...models import Category


class Command(BaseCommand):
    help = "Synchronizes categories"

    def handle(self, *args, **options):
        categories_to_sync = Category.objects.count()

        message = "Synchronizing %s categories...\n"
        self.stdout.write(message % categories_to_sync)

        message = "\n\nSynchronized %s categories in %s"

        start_time = time.time()

        synchronized_count = 0
        show_progress(self, synchronized_count, categories_to_sync)
        for category in Category.objects.iterator():
            category.synchronize()
            category.save()

            synchronized_count += 1
            show_progress(self, synchronized_count, categories_to_sync)

        end_time = time.time() - start_time
        total_time = time.strftime("%H:%M:%S", time.gmtime(end_time))

        self.stdout.write(message % (synchronized_count, total_time))



from django.urls import reverse

from ...core.utils import encode_json_html
from ..models import Ban
from ..test import UserTestCase


class DenyAuthenticatedTests(UserTestCase):
    def test_success(self):
        """deny_authenticated decorator allowed guest request"""
        response = self.client.post(reverse("misago:request-activation"))
        self.assertEqual(response.status_code, 200)

    def test_fail(self):
        """deny_authenticated decorator denied authenticated request"""
        self.login_user(self.get_authenticated_user())

        response = self.client.post(reverse("misago:request-activation"))
        self.assertEqual(response.status_code, 403)


class DenyGuestsTests(UserTestCase):
    def test_success(self):
        """deny_guests decorator allowed authenticated request"""
        self.login_user(self.get_authenticated_user())

        response = self.client.post(reverse("misago:options"))
        self.assertEqual(response.status_code, 200)

    def test_fail(self):
        """deny_guests decorator blocked guest request"""
        response = self.client.post(reverse("misago:options"))
        self.assertEqual(response.status_code, 403)

    def test_ref_login(self):
        """deny_guests decorator redirected guest request to homepage if ref=login"""
        response = self.client.post("%s?ref=login" % reverse("misago:options"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], reverse("misago:index"))


class DenyBannedIPTests(UserTestCase):
    def test_success(self):
        """deny_banned_ips decorator allowed unbanned request"""
        Ban.objects.create(
            check_type=Ban.IP, banned_value="83.*", user_message="Ya got banned!"
        )

        response = self.client.post(reverse("misago:request-activation"))
        self.assertEqual(response.status_code, 200)

    def test_fail(self):
        """deny_banned_ips decorator denied banned request"""
        Ban.objects.create(
            check_type=Ban.IP, banned_value="127.*", user_message="Ya got banned!"
        )

        response = self.client.post(reverse("misago:request-activation"))
        self.assertContains(
            response, encode_json_html("<p>Ya got banned!</p>"), status_code=403
        )



from django.urls import reverse

from ....test import assert_contains


def test_plugins_list_contains_plugins(admin_client):
    response = admin_client.get(reverse("misago:admin:plugins:index"))

    assert_contains(response, "Example plugin")
    assert_contains(response, "empty_manifest_plugin")
    assert_contains(response, "invalid_manifest_plugin")
    assert_contains(response, "minimal_plugin")


def test_plugins_list_contains_plugins_directories_and_packages(admin_client):
    response = admin_client.get(reverse("misago:admin:plugins:index"))

    assert_contains(response, "empty-manifest-plugin")
    assert_contains(response, "empty_manifest_plugin")
    assert_contains(response, "full-manifest-plugin")
    assert_contains(response, "full_manifest_plugin")
    assert_contains(response, "invalid-manifest-plugin")
    assert_contains(response, "invalid_manifest_plugin")
    assert_contains(response, "minimal-plugin")
    assert_contains(response, "minimal_plugin")


def test_plugins_list_contains_plugins_metadata(admin_client):
    response = admin_client.get(reverse("misago:admin:plugins:index"))

    assert_contains(response, "Rafał Pitoń")
    assert_contains(response, "0.1DEV")
    assert_contains(response, "GNU GPL v2")



from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponse
from rest_framework.decorators import api_view
from social_core.backends.github import GithubOAuth2
from social_core.exceptions import AuthFailed, NotAllowedToDisconnect, WrongBackend

from .. import errorpages
from ...users.models import Ban
from ..decorators import require_POST
from ..exceptions import Banned, SocialAuthBanned, SocialAuthFailed
from ..shortcuts import paginate, paginated_response, validate_slug
from ..views import home_redirect
from .models import Model
from .serializers import MockSerializer


def test_pagination(request, page=None):
    items = range(15)
    page = paginate(items, page, 5)
    return HttpResponse(",".join([str(x) for x in page.object_list]))


@api_view()
def test_paginated_response(request):
    data = range(100)
    page = paginate(data, 2, 10)

    return paginated_response(page)


@api_view()
def test_paginated_response_data(request):
    data = range(100)
    page = paginate(data, 2, 10)

    return paginated_response(page, data=["a", "b", "c", "d", "e"])


@api_view()
def test_paginated_response_serializer(request):
    data = [0, 1, 2, 3]
    page = paginate(data, 0, 10)

    return paginated_response(page, serializer=MockSerializer)


@api_view()
def test_paginated_response_data_serializer(request):
    data = [0, 1, 2, 3]
    page = paginate(data, 0, 10)

    return paginated_response(
        page, data=["a", "b", "c", "d"], serializer=MockSerializer
    )


@api_view()
def test_paginated_response_data_extra(request):
    data = [0, 1, 2, 3]
    page = paginate(data, 0, 10)

    return paginated_response(
        page, data=["a", "b", "c", "d"], extra={"next": "EXTRA", "lorem": "ipsum"}
    )


def validate_slug_view(request, pk, slug):
    model = Model(int(pk), "eric-the-fish")
    validate_slug(model, slug)
    return HttpResponse("Allright!")


def raise_misago_banned(request):
    ban = Ban(user_message="Banned for test!")
    raise Banned(ban)


def raise_misago_403(request):
    raise PermissionDenied("Misago 403")


def raise_misago_404(request):
    raise Http404("Misago 404")


def raise_misago_405(request):
    return errorpages.not_allowed(request)


def raise_403(request):
    raise PermissionDenied()


def raise_404(request):
    raise Http404()


def raise_social_auth_failed(require_POST):
    raise AuthFailed(GithubOAuth2)


def raise_social_wrong_backend(request):
    raise WrongBackend("facebook")


def raise_social_not_allowed_to_disconnect(request):
    raise NotAllowedToDisconnect()


def raise_social_auth_failed_message(request):
    raise SocialAuthFailed(GithubOAuth2, "This message will be shown to user!")


def raise_social_auth_banned(request):
    ban = Ban(user_message="Banned in auth!")
    raise SocialAuthBanned(GithubOAuth2, ban)


def test_redirect(request):
    return home_redirect(request)


@require_POST
def test_require_post(request):
    return HttpResponse("Request method: %s" % request.method)


@errorpages.shared_403_exception_handler
def mock_custom_403_error_page(request, exception):
    return HttpResponse("Custom 403", status=403)


@errorpages.shared_404_exception_handler
def mock_custom_404_error_page(request, exception):
    return HttpResponse("Custom 404", status=404)
